#pragma warning(disable:4786)
#include <windows.h>
#include <stdio.h>
#include "ComMallocSpyImpl.h"

//----------------------------------------------------------------------
KComMallocSpyImpl::KComMallocSpyImpl( DWORD Flags )
{
    m_bTraceOn = (Flags & ENABLE_TRACE) != 0;
    m_RefCounter = 1;
}

//----------------------------------------------------------------------
KComMallocSpyImpl::~KComMallocSpyImpl()
{
    // Nothing to do here
}

//----------------------------------------------------------------------
//  QueryInterface - Only IUnknown and IMallocSpy are meaningful

HRESULT KComMallocSpyImpl::QueryInterface(REFIID riid, LPVOID *ppUnk)
{
    HRESULT hr = S_OK;

    if (IsEqualIID(riid, IID_IUnknown))
    {
        *ppUnk = (IUnknown *) this;
        AddRef();
    }
    else if (IsEqualIID(riid, IID_IMallocSpy))
    {
        *ppUnk =  (IMallocSpy *) this;
        AddRef();
    }
    else
    {
        *ppUnk = NULL;
        hr =  E_NOINTERFACE;
    }

    return hr;
}

//----------------------------------------------------------------------
ULONG KComMallocSpyImpl::AddRef()
{
    // no reference counting - life time is governed by explicit new and delete
    return ++m_RefCounter; 
}

//----------------------------------------------------------------------
ULONG KComMallocSpyImpl::Release(void)
{
    // no reference counting - life time is governed by explicit new and delete
    if (--m_RefCounter == 0)
    {
        delete this;
        return 0;
    }

    return m_RefCounter;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PreAlloc
//
//  Synopsis:   Called prior to OLE calling IMalloc::Alloc
//
//  Arguments:  [cbRequest] -- The number of bytes the caller of
//                             is requesting IMalloc::Alloc
//
//  Returns:    The count of bytes to actuallay allocate
//
//----------------------------------------------------------------------
ULONG KComMallocSpyImpl::PreAlloc(ULONG cbRequest)
{
    m_nRequestedSize = cbRequest;
    return cbRequest;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PostAlloc
//
//  Synopsis:   Called after OLE calls IMalloc::Alloc
//
//  Arguments:  [pActual] -- The allocation returned by IMalloc::Alloc
//
//  Returns:    The allocation pointer to return to the caller of
//              IMalloc::Alloc
//
//----------------------------------------------------------------------
void *KComMallocSpyImpl::PostAlloc(void *pActual)
{
    // store allocated pointer in the map
    printf( "MALLOC SPY: 0x%08X - %u bytes allocated\n", pActual, m_nRequestedSize );

    m_Pointers.insert(CPointerMap::value_type(pActual, m_nRequestedSize));
    return pActual;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PreFree
//
//  Synopsis:   Called prior to OLE calling IMalloc::Free
//
//  Arguments:  [pRequest] -- The allocation to be freed
//              [fSpyed]   -- Whether it was allocated with a spy active
//
//  Returns:
//
//----------------------------------------------------------------------
void *KComMallocSpyImpl::PreFree(void *pRequest, BOOL fSpyed)
{
    // Use the stored information
    // ...

    // Check for NULL
    if (pRequest == NULL)
    {
        return NULL;
    }

    if (fSpyed)
    {
        if (m_bTraceOn)
        {
            printf("MALLOC SPY: 0x%08X - freed", pRequest );
        }

        int nRemoved = m_Pointers.erase(pRequest);

        if (m_bTraceOn)
        {
            if (nRemoved == 0)
            {
                printf(", NOT FOUND IN THE MAP");
            }
        };

        if (m_bTraceOn)
        {
            printf(", %u blocks remaining\n", m_Pointers.size());
        }
    };

    return pRequest;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PostFree
//
//  Synopsis:   Called after OLE calls IMalloc::Free
//
//  Arguments:  [fSpyed]   -- Whether it was allocated with a spy active
//
//----------------------------------------------------------------------
void KComMallocSpyImpl::PostFree(BOOL /*fSpyed*/)
{
    return;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PreRealloc
//
//  Synopsis:   Called prior to OLE calling IMalloc::Realloc
//
//  Arguments:  [pRequest]     -- The buffer to be reallocated
//              [cbRequest]    -- The requested new size of the buffer
//              [ppNewRequest] -- Where to store the new buffer pointer
//                                to be reallocated
//              [fSpyed]       -- Whether it was allocated with a spy active
//
//  Returns:    The new size to actually be allocated
//
//----------------------------------------------------------------------
ULONG KComMallocSpyImpl::PreRealloc(void *pRequest, ULONG cbRequest,
                             void **ppNewRequest, BOOL fSpyed)
{
    *ppNewRequest = pRequest;
    if (m_bTraceOn)
    {
        printf( "MALLOC SPY: 0x%08X - realloc to %u bytes\n", pRequest, cbRequest);
    }

    if (fSpyed)
    {
        m_Pointers.erase(pRequest);
    };
        
    return cbRequest;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PostRealloc
//
//  Synopsis:   Called after OLE calls IMalloc::Realloc
//
//  Arguments:  [pActual] -- Pointer to the reallocated buffer
//              [fSpyed]  -- Whether it was allocated with a spy active
//
//  Returns:    The buffer pointer to return
//
//----------------------------------------------------------------------
void *KComMallocSpyImpl::PostRealloc(void *pActual, BOOL fSpyed)
{
    // Return the buffer with the header offset
    if (fSpyed)
    {
        PostAlloc(pActual);
    };

    return pActual;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PreGetSize
//
//  Synopsis:   Called prior to OLE calling IMalloc::GetSize
//
//  Arguments:  [pRequest] -- The buffer whose size is to be returned
//              [fSpyed]   -- Whether it was allocated with a spy active
//
//  Returns:    The actual buffer with which to call IMalloc::GetSize
//
//----------------------------------------------------------------------
void *KComMallocSpyImpl::PreGetSize(void *pRequest, BOOL /*fSpyed*/)
{
    return pRequest;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PostGetSize
//
//  Synopsis:   Called after OLE calls IMalloc::GetSize
//
//  Arguments:  [cbActual] -- The result of IMalloc::GetSize
//              [fSpyed]   -- Whether it was allocated with a spy active
//
//  Returns:    The size to return to the IMalloc::GetSize caller
//
//----------------------------------------------------------------------
ULONG KComMallocSpyImpl::PostGetSize(ULONG cbActual, BOOL /*fSpyed*/)
{
    return cbActual;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PreDidAlloc
//
//  Synopsis:   Called prior to OLE calling IMalloc::DidAlloc
//
//  Arguments:  [pRequest] -- The buffer whose allocation is being tested
//              [fSpyed]   -- Whether it was allocated with a spy active
//
//  Returns:    The buffer whose allocation is actually to be tested
//
//----------------------------------------------------------------------
void *KComMallocSpyImpl::PreDidAlloc(void *pRequest, BOOL /*fSpyed*/)
{
    return pRequest;
}

//+---------------------------------------------------------------------
//
//  Function:   PostDidAlloc
//
//  Synopsis:   Called after OLE calls the IMalloc::DidAlloc
//
//  Arguments:  [pRequest] -- The passed allocation
//              [fSpyed]   -- Whether it was allocated with a spy active
//              [fActual]  -- The result of IMalloc::DidAlloc
//
//  Returns:    The result of IMalloc::DidAlloc
//
//----------------------------------------------------------------------
BOOL KComMallocSpyImpl::PostDidAlloc(void* /*pRequest*/, BOOL /*fSpyed*/, BOOL fActual)
{
    return fActual;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PreHeapMinimize
//
//  Synopsis:   Called prior to OLE calling the IMalloc::HeapMinimize
//
//----------------------------------------------------------------------
void KComMallocSpyImpl::PreHeapMinimize(void)
{
    // We don't do anything here
    return;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::PostHeapMinimize
//
//  Synopsis:   Called after OLE calls the IMalloc::HeapMinimize
//
//----------------------------------------------------------------------
void KComMallocSpyImpl::PostHeapMinimize(void)
{
    // We don't do anything here
    return;
}

//+---------------------------------------------------------------------
//
//  Member:     KComMallocSpyImpl::CheckLeaks
//
//  Synopsis:   Checks memory leaks - any not freed data
//
//  NOTE:       TRACE() works only in Debug. In release it is no-op
//              OutputDebugString() works always
//
//----------------------------------------------------------------------
void KComMallocSpyImpl::CheckLeaks()
{
    if (!m_Pointers.empty())
    {
        printf("MALLOC SPY: detected memory leaks!\n");

        for (CPointerMap::iterator P = m_Pointers.begin(); P != m_Pointers.end(); ++P)
        {
            printf("0x%08X - %u bytes: ", P->first, P->second);
            int Size = min(P->second, 16);

            try
            {
                BYTE* Ptr = (BYTE*)P->first;
                for (int i=0; i<Size; ++i)
                {
                    printf("%02X ", Ptr[i]);
                };

                for (i=0; i<Size; ++i)
                {
                    if (isprint(Ptr[i]))
                        printf("%c", Ptr[i] );
                    else
                        printf(".");
                };
            }
            catch (...)
            {
            }

            printf("\n");
        }
    };
};
